/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <gnome.h>
#include <gtk/gtk.h>
#include <pthread.h>
#include <errno.h>

#include "../uulib/uudeview.h"
#include "../uulib/fptools.h"

#include "pan.h"

static int must_exit = FALSE;
static int stop_all = FALSE;
static int online = FALSE;
static pthread_cond_t online_c = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t online_m = PTHREAD_MUTEX_INITIALIZER;

/**************************************************
* Cleanup Handler Stuff
***************************************************/

void (*cleanup) _ANSI_ARGS_((void *, int)) = NULL;
void *cleanup_arg = NULL;

void set_cleanup(void *opaque, void (*func) _ANSI_ARGS_((void *, int)))
{
	cleanup = func;
	cleanup_arg = opaque;
}

/* status = 2 when errors interrupt 
 * status = 1 when operation was interrupted
 * status = 0 when finished */

void run_cleanup(int status)
{
	if (cleanup && cleanup_arg)
		(*cleanup) (cleanup_arg, status);
	cleanup = NULL;
	cleanup_arg = NULL;
}

/**************************************************
* go on/off line code
***************************************************/

void connect_dialog_clicked_cb(GnomeDialog *dialog, int button, int fd[2])
{
	char *p;
	int i;
	
	switch (button) {
	case 0:					/* OK button */
		i = 1;
		break;
	default:			/* Cancel Button */
		i = 0;
		break;
	}
	write(fd[1], &i, sizeof(int));
	gtk_widget_destroy(GTK_WIDGET(dialog));
}

void connect_wait (void)
{              
	pthread_mutex_lock(&online_m);
	pthread_cond_wait(&online_c, &online_m);
	pthread_mutex_unlock(&online_m);
}

int go_online (void)
{
	if (online == FALSE) {
		if (connect_to_server() == 0) {
			online = TRUE;
			gui_set_status_default(_("Pan - Online"));
			gui_clean();
			log_add("Pan - Online");
			pthread_mutex_lock(&online_m);
			pthread_cond_broadcast(&online_c);
			pthread_mutex_unlock(&online_m);
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		return TRUE;
	}
}

int go_online_ask(void)
{
	GtkWidget *connect_dialog;
	GtkWidget *label;
	static int fd[2];
	int i = 0;

	if (online == FALSE) {
		/* We setup a pipe for communication between the threads */
		if (pipe(fd) != 0) {
			g_warning("Couldn't create pipe");
			return FALSE;
		}
		/* Spawn the go_online? dialog, make it modal so we don't
	 	* leave the network thread in limbo forever */
		pan_lock();
		connect_dialog = gnome_dialog_new(_("Go online?"), GNOME_STOCK_BUTTON_YES, GNOME_STOCK_BUTTON_NO, NULL);
		label = gtk_label_new(_("A connection to the server is needed.\nShall we? "));
		gtk_container_add(GTK_CONTAINER(GNOME_DIALOG(connect_dialog)->vbox), label);
	   	gtk_signal_connect(GTK_OBJECT(connect_dialog), "clicked", GTK_SIGNAL_FUNC(connect_dialog_clicked_cb), fd);
		gtk_window_set_modal(GTK_WINDOW(connect_dialog), TRUE);
		gtk_widget_show_all(connect_dialog);
		pan_unlock();

		read(fd[0], &i, sizeof(int));
		if (i == 1) {
			return go_online();
		} else {
			return FALSE;
		}
	} else {
		return TRUE;
	}
}

void go_offline(void)
{
	if (online == TRUE) {
		put_server("QUIT\r\n");
		tcp_close();
		must_exit = FALSE;
		gui_set_status_default(_("Pan - Offline"));
		log_add("Pan - Offline");
		gui_clean();
		if (stop_all) {
			queue_purge();
			stop_all = FALSE;
		}
		online = FALSE;
	}
}

int is_online(void)
{
	return online;
}

void network_error(const char *format, ...)
{
	va_list args;
	char *p = NULL;

	g_return_if_fail(format != NULL);

	va_start(args, format);
	p = g_strdup_vprintf(format, args);
	va_end(args);

	pan_lock();
	gnome_error_dialog(p);
	pan_unlock();
	g_free(p);
	/* Run any cleanup with status=2 (error) */
	run_cleanup(2);
	go_offline();
}

int set_group(group_data * data)
{
	char *buffer;

	g_assert(data);

	if (put_server_va("GROUP %s\r\n", data->name) == -1) {
		return 0;
	}
	if ((buffer = get_server()) == NULL) {
		return 0;
	}
	if (atoi(buffer) != 211)
		return 1;
	else
		return 0;
}


/***************************************************************
* article_post
****************************************************************/

void article_post(queue_data * qdata)
{
	char *buffer;
	char *message;

	message = qdata->str;

	gui_set_status(_("Posting message..."));
	if (put_server("POST\r\n") == -1) {
		network_error(_("Error writing to socket during post."));
		return;
	}
	if ( (buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket during post."));
		return;
	}
	if (atoi(buffer) == 340) {
		if (put_server_va("%s\r\n.\r\n", message) == -1) {
			network_error(_("Error writing to socket during post."));
			return;
		}
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket during post."));
			return;
		}
		if (atoi(buffer) == 240) {
			gui_set_info(_("Message posted."));
			gui_clean();
			return;
		}
	}
	gui_clean();
	pan_error_dialog(_("Error posting message.  Server said:\n'%s'"), buffer);
}


/***************************************************************
* grouplist_download 
****************************************************************/

void grouplist_download_cleanup(glist_c * c, int status)
{
	switch (status) {
	case 1:					/* Interrupt */
	case 2:					/* Error */
		grouplist_free(c->glist);
		break;
	default:					/* Success */
		break;
	}
	g_free(c);
}

void grouplist_download(void)
{
	char *buffer = NULL;
	group_data *gdata = NULL;
	int count = 0;
	glist_c *c;

	c = g_malloc0(sizeof(glist_c));
	set_cleanup(c, grouplist_download_cleanup);

	grouplist_free(Pan.grouplist);
	Pan.grouplist = NULL;
	gui_set_status(_("Retrieved 0 groups."));

	if (put_server("LIST active\r\n") == -1) {
		network_error(_("Error writing to socket."));
		return;
	}
	if ( (buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	if (atoi(buffer) != 215) {
		network_error(_("Group list could not be retrieved\nServer said: %s"), buffer);
		return;
	} else {
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}
		gui_set_progress_activity(1);
		while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) {
				run_cleanup(1);
				go_offline();
				return;
			}
			gdata = (group_data *) g_malloc0(sizeof(group_data));
			gdata->name = g_strdup(strtok(buffer, " "));
			c->glist = g_slist_prepend(c->glist, gdata);
			gui_set_progress_activity(++count);
			if ((count % 23) == 0) {
				gui_set_status_va(_("Retrieved %d groups."), count);
			}
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				return;
			}
		}
	}
	gui_set_info_va(_("Retrieved %d groups."), count);
	log_add_va("Retrieved %d groups", count);
	gui_set_status_va(_("Sorting %d groups."), count);
	c->glist = g_slist_sort(c->glist, (GCompareFunc) compare_group_data);
	Pan.grouplist = c->glist;
	grouplist_save();
	grouplist_update();
	run_cleanup(0);
}

/***************************************************************
* articlelist_download 
****************************************************************/

void articlelist_download_cleanup(alist_c * c, int status)
{
	switch (status) {
	case 2:					/* Error */
	case 1:					/* Interrupt */
		articlelist_free(c->alist);
		break;
	default:					/* Success */
		break;
	}
	g_free(c);
}

void articlelist_download(queue_data * qdata)
{
	char *buffer = NULL;
	article_data *adata = NULL;
	int count = 0, total;
	int first = 0;
	int last = 0;
	int p_last = 0;
	int part = 0;
	int parts = 0;
	int i;
	group_data *gdata = NULL;
	time_t current_time, date;
	char **sd;
	int flags = qdata->flags;
	alist_c *c;
	char *s;
	int expiration;

	expiration = gnome_config_get_int_with_default("/Pan/Articles/Expiration=30", NULL);

	c = g_malloc0(sizeof(alist_c));
	set_cleanup(c, articlelist_download_cleanup);
	gdata = qdata->gdata;
	g_free(qdata);

	if (put_server_va("GROUP %s\r\n", gdata->name) == -1) {
		network_error(_("Error writing to socket."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}

	if (strtol(buffer, NULL, 10) != 211) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return;
	}

	sscanf(buffer, "%*d %d %d %d", &total, &first, &last);

	p_last = group_get_attrib_i(gdata, "Last");
	if (total == '0') {
		gui_set_info(_("No articles"));
		run_cleanup(0);
		return;
	}
	if ((flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) && p_last) {
		if (p_last == last) {
			gui_set_info(_("No new articles"));
			run_cleanup(0);
			return;
		}
		gui_set_status(_("Retrieving New Article Headers"));
		if (put_server_va("XOVER %d-%d\r\n", p_last, last) == -1) {
			network_error(_("Error writing to socket."));
			return;
		}
	} else {
		if ((total > 150) && ((flags & 0x003F) == QUEUE_ARTICLELIST_DOWNLOAD_SAMPLE)) {
			gui_set_status(_("Retrieving Sample of Article Headers"));
			if (put_server_va("XOVER %d-%d\r\n", first, first + 150) == -1) {
				network_error(_("Error writing to socket."));
				return;
			}
		} else {
			gui_set_status(_("Retrieving All Article Headers"));
			if (put_server_va("XOVER %d-%d\r\n", first, last) == -1) {
				network_error(_("Error writing to socket."));
				return;
			}
		}
	}

	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	if (strtol(buffer, (char **) NULL, 10) != 224) {
		network_error(_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return;
	}
	time(&current_time);
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		sd = g_strsplit(buffer, "\t", 0);
		date = parse_date(sd[3]);
		if ((current_time - date) > (86400 * expiration)) {
			g_strfreev(sd);
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				return;
			}
			continue;
		}
		adata = (article_data *) g_malloc0(sizeof(article_data));
		adata->number = g_strdup(sd[0]);
		adata->subject = g_strdup(sd[1]);
		adata->author = g_strdup(sd[2]);
		adata->date = date;
		adata->message_id = g_strdup(sd[4]);
		adata->references = g_strdup(sd[5]);
		adata->linecount = strtol(sd[7], NULL, 10);
		adata->group = gdata;
		adata->state = STATE_NEW;

		g_strfreev(sd);

		/* Here we look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = adata->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit(s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit(s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* Verify Multipart info */
		if ((parts > 1) && (part <= parts)) {
			/* We found one! */
			adata->parts = parts;
			adata->part = part;
		} else {
			adata->parts = 1;
			adata->part = 1;
		}
		/* Add the article to the article list */
		c->alist = g_slist_prepend(c->alist, adata);

		gui_set_progress((float)++ count / (float) total);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}
	}
	/* Check to see if all articles were expired */
	if (c->alist == NULL) {
		gui_set_info(_("All articles were expired"));
		run_cleanup(0);
		return;
	}
	group_set_attrib_i(gdata, "Last", last);

	if (count > 0) {
		gui_set_info_va(_("Retrieved %d article headers"), count);
		log_add_va("Retrieved %d article headers from %s", count, gdata->name);

		if (flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) {
			/* Add new headers to existing */
			if (gdata != Pan.active_group) {
				gdata->articlelist = db_load_articles(gdata);
			} else {
				/* Take the new flag off the old list, not done above because db_load_articles does
				 * it automatically */
				articlelist_not_new(gdata->articlelist);
			}
			gdata->articlelist = g_slist_concat(gdata->articlelist, c->alist);
			db_save_articles(gdata, c->alist, count);	
			group_add_attrib_i(gdata, "Unread", count);
			group_add_attrib_i(gdata, "Total", count);
		} else {
			/* Save all the fresh new headers */
			gdata->articlelist = c->alist;
			db_save_articles(gdata, gdata->articlelist, count);
			group_set_attrib_i(gdata, "Unread", count);
			group_set_attrib_i(gdata, "Total", count);
		}
		/* Now we thread the articles */
		gui_set_status(_("Threading Article List"));
		gui_set_progress(0);
		gdata->articlelist = thread_list_by_multipart(gdata->articlelist, gdata);
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		
		gdata->articlelist = thread_list_by_subject(gdata->articlelist, gdata);
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		
		gdata->articlelist = thread_list_by_reference(gdata->articlelist, gdata);
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		
		grouplist_update_row(gdata);
	
		if (flags & QUEUE_FOREGROUND) {
			gui_set_status(_("Displaying article headers"));
			articlelist_update((group_data *) gdata);
		}
	} else {
		gui_set_info(_("No new articles"));
	}
			
	run_cleanup(0);
}

/***************************************************************
* article_download 
***************************************************************/

void article_download_cleanup(ad_c * c, int status)
{
	if (c->adata) {
		c->adata->state &= ~STATE_DOWNLOAD_QUEUED;
		if (status == 2)
			c->adata->state |= STATE_ERROR;
		else
			c->adata->state &= ~STATE_ERROR;
		article_update(c->adata);
		article_free(c->adata);
	}
	g_free(c->qdata);
	g_free(c);
}

void
article_download (queue_data* qdata)
{
	char *buffer;
	group_data *gdata;
	int count;
	ad_c *c;

	c = g_malloc0(sizeof(ad_c));
	set_cleanup(c, article_download_cleanup);
	gdata = qdata->gdata;
	c->qdata = qdata;

	if ((c->adata = db_fetch_article(gdata, qdata->str)) == NULL) {
		gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
		run_cleanup(2);
		return;
	}
	gui_set_status_va(_("Downloading article '%s'"), c->adata->subject);

	if (set_group(gdata)) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return;
	}

	if (put_server_va("HEAD %s\r\n", c->adata->number) == -1) {
		network_error(_("Error writing to socket."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}

	if (strtol(buffer, (char **) NULL, 10) != 221) {
		run_cleanup(0);
		return;
	}
	c->adata->header = g_strdup("");
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			run_cleanup(2);
			go_offline();
			return;
		}
		c->adata->header = g_realloc(c->adata->header, strlen(c->adata->header) + strlen(buffer) + 1);
		c->adata->header = strcat(c->adata->header, buffer);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}
	}

	if (put_server_va("BODY %s\r\n", c->adata->number) == -1) {
		network_error(_("Error writing to socket."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	if (strtol(buffer, (char **) NULL, 10) != 222) {
		run_cleanup(2);
		return;
	}
	c->adata->body = g_strdup("");
	count = 0;
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		c->adata->body = g_realloc(c->adata->body, strlen(c->adata->body) + strlen(buffer) + 1);
		c->adata->body = strcat(c->adata->body, buffer);
		gui_set_progress((float) count++ / (float) c->adata->linecount);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}
	}
	gui_set_progress(1);

	if (qdata->flags & QUEUE_FOREGROUND)
		text_update ((article_data *) c->adata);

	if ( (c->adata->state == STATE_NONE) ||
	     (c->adata->state == STATE_NEW) )
		group_add_attrib_i(gdata, "Unread", -1);
	c->adata->state |= STATE_READ;
	run_cleanup(0);
}


/***************************************************************
* UUdeview Lib Stuff 
***************************************************************/

void UUD_message_callback(void *opaque, char *msg, int level)
{
	if ((level == UUMSG_WARNING) || (level == UUMSG_ERROR)) {
		pan_error_dialog(_("An error occured during decoding.\n%s"), msg);
	}
}

int UUD_busy_callback(void *opaque, uuprogress * progress)
{
	float t = (float) 1 / (float) 3;

	switch (progress->action) {
	case UUACT_SCANNING:
		gui_set_progress((float) progress->percent / (float) 100 * t);
		return (0);
	case UUACT_DECODING:
		gui_set_progress(((float) progress->percent / (float) 100 * t) +
				 t);
		return (0);
	case UUACT_COPYING:
		gui_set_progress(((float) progress->percent / (float) 100 * t) +
				 t + t);
		return (0);
	default:
		g_warning("Reached default in UUD_busy_callback\n");
		return (0);
	}
}

char *UUD_file_name_filter(void *opaque, char *fname)
{
	char *ptr;

	if (fname == NULL)
		return NULL;

	/*
	 * strip directory information
	 */

	if ((ptr = _FP_strrchr(fname, '/')) != NULL)
		return ptr + 1;
	else if ((ptr = _FP_strrchr(fname, '\\')) != NULL)
		return ptr + 1;

	return fname;
}

void decode(article_data * adata, char *filename)
{
	uulist *item;

	chdir(props_get_download_dir());
	UUInitialize();
	UUSetMsgCallback(adata, UUD_message_callback);
	UUSetBusyCallback(adata, UUD_busy_callback, 1);
	UUSetFNameFilter(adata, UUD_file_name_filter);
	if ((UULoadFile(filename, NULL, 0)) != UURET_OK) {
		pan_error_dialog(_("Decoded couldn't load temporary file:\n%s"), filename);
	} else if ((item = UUGetFileListItem(0)) != NULL) {
		if ((item->state & UUFILE_OK) != 0) {
			if ((UUDecodeFile(item, NULL)) != UURET_OK) {
				pan_error_dialog(_("Error decoding '%s'"), adata->subject);
				gui_set_info_va(_("Error decoding '%s'"), adata->subject);
			} else {
				adata->state |= STATE_DECODED;
				gui_set_info_va(_("Successfully decoded '%s'"), UUFNameFilter(item->filename));
				log_add_va("Successfully decoded '%s'", UUFNameFilter(item->filename));
			}
		} else {
			pan_error_dialog(_("Error decoding '%s'"), adata->subject);
			gui_set_info_va(_("Error decoding '%s'"), adata->subject);
		}
	}
	UUCleanUp();
}


/***************************************************************
* article_decode 
***************************************************************/

void article_decode_cleanup(adec_c * c, int status)
{
	if (c->adata) {
		c->adata->state &= ~STATE_DECODE_QUEUED;
		if (status == 2)		/* Error */
			c->adata->state |= STATE_ERROR;
		else
			c->adata->state &= ~STATE_ERROR;
		article_update(c->adata);
		article_free(c->adata);
	}
	if (c->filename) {
		if (c->tmp) {
			fclose(c->tmp);
		}
		unlink(c->filename);
		g_free(c->filename);
	}
	g_free(c->qdata);
	g_free(c);
}

void article_decode(queue_data * qdata)
{
	char *buffer = NULL;
	group_data *gdata = NULL;
	int count;
	adec_c *c;

	c = g_malloc0(sizeof(adec_c));
	set_cleanup(c, article_decode_cleanup);
	gdata = qdata->gdata;
	c->qdata = qdata;

	if ((c->adata = db_fetch_article(gdata, qdata->str)) == NULL) {
		g_warning("Couldn't find article header for message_id: %s", qdata->str);
		gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
		run_cleanup(2);
		return;
	}
	c->filename = tempnam(props_get_temp_dir(), "pan");
	c->tmp = fopen(c->filename, "w");

	if (c->tmp == NULL) {
		network_error(_("Couldn't open temporary file %s"), c->filename);
		return;
	}

	if (set_group(gdata)) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return;
	}

	gui_set_status_va(_("Downloading article body '%s'"), c->adata->subject);
	gui_set_progress(0);

	if (put_server_va("BODY %s\r\n", c->adata->number) == -1) {
		network_error(_("Error writing to socket."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	if (strtol(buffer, NULL, 10) != 222) {
		run_cleanup(2);
		return;
	}
	count = 1;
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket."));
		return;
	}
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) {
			run_cleanup(1);
			go_offline();
			return;
		}
		fputs(buffer, c->tmp);
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}
		gui_set_progress((float)count++/(float)c->adata->linecount);
	}
	gui_set_progress(1);
	fclose(c->tmp);
	c->tmp = NULL;

	gui_set_status_va(_("Decoding article: %s"), c->adata->subject);
	gui_set_progress(0);

	if ( (c->adata->state == STATE_NONE) ||
	     (c->adata->state == STATE_NEW) )
		group_add_attrib_i(gdata, "Unread", -1);
	decode(c->adata, c->filename);

	run_cleanup(0);
}



/***************************************************************
* multipart_decode 
***************************************************************/
void multipart_decode_cleanup(mdec_c * c, int status)
{
	if (c->fadata) {
		c->fadata->state &= ~STATE_DECODE_QUEUED;
		if (status == 2)		/* Error */
			c->fadata->state |= STATE_ERROR;
		else
			c->fadata->state &= ~STATE_ERROR;
		article_update(c->fadata);
		article_free(c->fadata);
	}
	if (c->adata)
		article_free(c->adata);
	if (c->filename) {
		if (c->tmp) {
			fclose(c->tmp);
		}
		unlink(c->filename);
		g_free(c->filename);
	}
	g_strfreev(c->sd);
	g_free(c->qdata);
	g_free(c);
}

void multipart_decode(queue_data * qdata)
{
	char *buffer;
	group_data *gdata;
	int i = 0;
	float count;
	mdec_c *c;

	c = g_malloc0(sizeof(mdec_c));
	set_cleanup(c, multipart_decode_cleanup);
	gdata = qdata->gdata;
	c->qdata = qdata;

	c->sd = g_strsplit(qdata->str, "\n", -1);
	g_assert(c->sd);

	/* Find the (fake) article header for the
	 * multipart node */
	if ((c->fadata = db_fetch_article(gdata, c->sd[2])) == NULL) {
		gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
		run_cleanup(2);
		return;
	}
	c->filename = tempnam(props_get_temp_dir(), "pan");

	if ((c->tmp = fopen(c->filename, "w")) == NULL) {
		gui_set_info(_("Multipart decode failed: Couldn't open tempfile"));
		run_cleanup(2);
		return;
	}
	if (set_group(gdata)) {
		network_error(_("Couldn't select group '%s'"), gdata->name);
		return;
	}
	for (i = 2; c->sd[i]; i++) {
		if ((c->adata = db_fetch_article(gdata, c->sd[i])) == NULL) {
			gui_set_info_va(_("Couldn't find article header for message_id: %s"), c->sd[i]);
			run_cleanup(2);
			return;
		}
		gui_set_status_va(_("Downloading part %d of %d: %s"),
				  c->adata->part, c->adata->parts,
				  c->adata->subject);
		gui_set_progress(0);

		if (put_server_va("BODY %s\r\n", c->adata->number) == -1) {
			network_error(_("Error writing to from socket."));
			return;
		}
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}

		if (strtol(buffer, (char **) NULL, 10) != 222) {
			run_cleanup(2);
			return;
		}
		count = 1;
		if ((buffer = get_server()) == NULL) {
			network_error(_("Error reading from socket."));
			return;
		}

		while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) {
				run_cleanup(1);
				return;
			}
			fputs(buffer, c->tmp);
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket."));
				return;
			}
			gui_set_progress(count++ / (float) c->adata->linecount);
		}
		article_free(c->adata);
		c->adata = NULL;
	}
	fclose(c->tmp);
	c->tmp = NULL;

	gui_set_progress(1);
	gui_set_status_va(_("Decoding multipart '%s'"), c->sd[1]);
	gui_set_progress(0);

	if ( (c->fadata->state == STATE_NONE) ||
	     (c->fadata->state == STATE_NEW) )
		group_add_attrib_i(gdata, "Unread", -1);
	decode(c->fadata, c->filename);

	run_cleanup(0);
}

void nntp_send_noop()
{
	char *buffer;

	if (put_server("HELP\r\n") == -1) {
		network_error(_("Error writing to socket during HELP no-op."));
		return;
	}
	if ((buffer = get_server()) == NULL) {
		network_error(_("Error reading from socket during HELP no-op."));
		return;
	}
	if (atoi(buffer) == 100) {
 		while (strncmp(buffer, ".\r\n", 3)) {
			if ((buffer = get_server()) == NULL) {
				network_error(_("Error reading from socket during HELP no-op."));
				return;
			}
		}
	}
}
		
int connect_to_server(void)
{
	char *buffer;

	if (Pan.active_server == NULL) {
		pan_error_dialog(_("There is no active server.\nPlease configure one."));
		return 1;
	}
	gui_set_status_va(_("Connecting to server %s"), Pan.active_server->name);

	if (tcp_open(Pan.active_server, "nntp") < 0) {
		gui_clean();
		pan_error_dialog(_("Could not connect to server."));
		return 1;
	}
	if ((buffer = get_server()) == NULL) {
		gui_clean();
		pan_error_dialog(_("Could not get response from server."));
		return 1;
	}
	/* Now we attempt authentication.  This is not implemented 
	 * exactly as "Common NNTP Extensions" specifies, being that
	 * we didn't wait to be asked for it, but hopefully it 
	 * will work. */
	if ((strlen(Pan.active_server->username) + strlen(Pan.active_server->password)) > 1) {
		if (put_server_va("AUTHINFO user %s\r\n", Pan.active_server->username) == -1) {
			pan_error_dialog(_("Error writing to socket."));
			return 1;
		}
		if ((buffer = get_server()) == NULL) {
			pan_error_dialog(_("Error reading from socket."));
			return 1;
		}
		if (atoi(buffer) == 381) {
			/* 381 Means: More authentication information required */
			if (put_server_va("AUTHINFO pass %s\r\n", Pan.active_server->password) == -1) {
				pan_error_dialog(_("Error writing to socket."));
				return 1;
			}
			if ((buffer = get_server()) == NULL) {
				pan_error_dialog(_("Error reading from socket."));
				return 1;
			}
			if (atoi(buffer) != 281) {
				/* 281 Means: Authentication accepted */
				gui_clean();
				gui_set_info_va(_("Authentication unsuccessful."), buffer);
				pan_error_dialog(_("Authentication unsuccessful.\nServer said: %s"), buffer);
				return 1;
			}
		}
		/* Else authentication not required or not understood */
	}
	if (put_server("MODE READER\r\n") == -1) {
		pan_error_dialog(_("Error writing to socket."));
		return 1;
	}
	if ((buffer = get_server()) == NULL) {
		pan_error_dialog(_("Error reading from socket."));
		return 1;
	}
	if (atoi(buffer) == 200)
		Pan.active_server->posting = 1;
	else if (atoi(buffer) == 201)
		Pan.active_server->posting = 0;
	else {
		gui_clean();
		pan_error_dialog(_("Could not connect to server.\nServer said: %s"), buffer);
		return (1);
	}
	return (0);
}

/*
* Menu callback functions 
*/
void go_offline_mf(void)
{
	if (is_online())
		must_exit = TRUE;
}

void stop_all_mf(void)
{
	if (is_online()) {
		must_exit = TRUE;
		stop_all = TRUE;
	}
}
